{ File   : MakeDAT.dpr
  Author : Deniz Oezmen
  Created: 2007-12-14
  Changed: 2007-12-14

  Dreamfall localization.dat creator.
}
program MakeDAT;

{$APPTYPE CONSOLE}

uses
  SysUtils,
  StrUtils;

type
  TLangHeader  = record
                   AbbrevLen: Longint;
                   Abbrev   : string;
                   NameLen  : Longint;
                   Name     : string;
                   NumLines,
                   Size     : Longint;
                 end;

  TLangLine    = packed record
                   ID,
                   Offset: Longint;
                 end;

  TLangLines = array of TLangLine;

var
  TxtFile: TextFile;
  DATFile: file;

function OpenTxtFile(const TxtName: string; var TxtFile: TextFile): Boolean;
begin
  AssignFile(TxtFile, TxtName);
  {$I-}
  Reset(TxtFile);
  {$I+}
  Result := (IOResult = 0)
end;

function OpenDATFile(const DATName: string; var DATFile: file): Boolean;
begin
  AssignFile(DATFile, DATName);
  {$I-}
  Rewrite(DATFile, 1);
  {$I+}
  Result := (IOResult = 0)
end;

procedure WriteDAT(var TxtFile: TextFile; var DATFile: file);
var
  i, p,
  BaseOfs    : Longint;
  Line       : string;
  LangHeader : TLangHeader;
  LangLines  : TLangLines;
  LangNums,
  LangOffsets: array of Longint;
  LangStrings: array of string;
begin
  while not Eof(TxtFile) do
  begin
    // read abbrev
    ReadLn(TxtFile, Line);
    if LeftStr(Line, 7) <> 'abbrev=' then
    begin
      WriteLn(Format('"abbrev=" expected, found "%s"', [Line]));
      Exit
    end;
    with LangHeader do
    begin
      AbbrevLen := Length(Line) - 7;
      Abbrev := RightStr(Line, AbbrevLen);
    end;

    // read name
    ReadLn(TxtFile, Line);
    if LeftStr(Line, 5) <> 'name=' then
    begin
      WriteLn(Format('"name=" expected, found "%s"', [Line]));
      Exit
    end;
    with LangHeader do
    begin
      NameLen := Length(Line) - 5;
      Name := RightStr(Line, NameLen);
    end;

    // read numlines
    ReadLn(TxtFile, Line);
    if LeftStr(Line, 9) <> 'numlines=' then
    begin
      WriteLn(Format('"numlines=" expected, found "%s"', [Line]));
      Exit
    end;
    SetLength(LangLines, StrToInt(RightStr(Line, Length(Line) - 9)));
    SetLength(LangNums, Length(LangLines));

    // read numstrings
    ReadLn(TxtFile, Line);
    if LeftStr(Line, 11) <> 'numstrings=' then
    begin
      WriteLn(Format('"numstrings=" expected, found "%s"', [Line]));
      Exit
    end;
    SetLength(LangStrings, StrToInt(RightStr(Line, Length(Line) - 11)));
    SetLength(LangOffsets, Length(LangStrings));

    for i := 0 to High(LangLines) do
    begin
      ReadLn(TxtFile, Line);

      p := Pos('=', Line);

      LangLines[i].ID := StrToInt(LeftStr(Line, p - 1));
      LangNums[i] := StrToInt(RightStr(Line, Length(Line) - p))
    end;

    BaseOfs := 0;
    for i := 0 to High(LangStrings) do
    begin
      ReadLn(TxtFile, Line);

      Line := AnsiReplaceStr(Line, '\n', #013#010);
      Line := AnsiReplaceStr(Line, '\d', #013);
      Line := AnsiReplaceStr(Line, '\a', #010);

      LangStrings[i] := Line  + #000;
      LangOffsets[i] := BaseOfs;
      Inc(BaseOfs, Length(LangStrings[i]))
    end;

    with LangHeader do
    begin
      NumLines := Length(LangLines);
      Size := BaseOfs
    end;

    for i := 0 to High(LangLines) do
      LangLines[i].Offset := LangOffsets[LangNums[i]];

    with LangHeader do
    begin
      BlockWrite(DATFile, AbbrevLen, 4);
      BlockWrite(DATFile, Abbrev[1], AbbrevLen);
      BlockWrite(DATFile, NameLen, 4);
      BlockWrite(DATFile, Name[1], NameLen);
      BlockWrite(DATFile, NumLines, 4);
      BlockWrite(DATFile, Size, 4)
    end;

    BlockWrite(DATFile, LangLines[0],
      Length(LangLines) * SizeOf(TLangLine));

    for i := 0 to High(LangStrings) do
      BlockWrite(DATFile, LangStrings[i][1], Length(LangStrings[i]))
  end
end;

begin
  if ParamCount <> 2 then
  begin
    WriteLn('Incorrect number of parameters');
    Exit
  end;

  if not OpenTxtFile(ParamStr(1), TxtFile) then
  begin
    WriteLn('Could not open input file.');
    Exit
  end;

  if not OpenDATFile(ParamStr(2), DATFile) then
  begin
    WriteLn('Could not write to output file.');
    CloseFile(TxtFile);
    Exit
  end;

  WriteDAT(TxtFile, DATFile);

  CloseFile(TxtFile);
  CloseFile(DATFile)
end.

