{ File   : PAKExt.dpr
  Author : Deniz Oezmen
  Created: 2006-09-09
  Changed: 2010-03-22

  Extractor for Dreamfall PAK archives.
}
program PAKExt;

{$APPTYPE CONSOLE}

uses
  SysUtils, StrUtils;

const
  NameTableInv: array[0..255] of Byte =
    ( 0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
      0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
      0,  0,  0,  0,  0,  0,  0, 32,  0,  0,  0,  0,  0, 30, 33, 27,
     34, 35, 36, 37, 38, 39, 40, 41, 42, 43,  0,  0,  0,  0,  0,  0,
      0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14, 15,
     16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26,  0,  0,  0,  0, 31,
      0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14, 15,
     16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26,  0,  0,  0,  0,  0,
      0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
      0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
      0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
      0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
      0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
      0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
      0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
      0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0);

type
  TPAKHeader  = packed record
                  ID          : array[0..11] of Char;
                  NumHashNodes,
                  NumNameIdxs,
                  NumNames    : Longint
                end;

  THashNode  = packed record
                 Offset,
                 Size,
                 BaseNodeIdx,
                 PathLen,
                 NameIdx    : Longint
               end;

  THashNodes = array of THashNode;

  TNames     = array of Byte;

  TNameIdxs  = array of Longint;

var
  HashNodes: THashNodes;
  Names    : TNames;
  NameIdxs : TNameIdxs;  

function OpenPAKFile(var PAKFile: file; const PAKName: string): Boolean;
begin
  FileMode := fmShareDenyWrite;
  AssignFile(PAKFile, PAKName);

  {$I-}
  Reset(PAKFile, 1);
  {$I+}
  OpenPAKFile := (IOResult = 0)
end;

function OpenDirFile(var DirFile: TextFile; const DirName: string): Boolean;
begin
  FileMode := fmShareDenyWrite;
  AssignFile(DirFile, DirName);

  {$I-}
  Reset(DirFile);
  {$I+}
  OpenDirFile := (IOResult = 0)
end;

function ReadHeader(var PAKFile: file; out PAKHeader: TPAKHeader): Boolean;
var
  BytesRead: Longword;
begin
  BlockRead(PAKFile, PAKHeader, SizeOf(TPAKHeader), BytesRead);
  ReadHeader := (SizeOf(TPAKHeader) = BytesRead) and
    (PAKHeader.ID = 'tlj_pack0001')
end;

function MakeDir(const FileName: string): Boolean;
var
  Dir,
  SubDir: string;
  p     : Longint;
begin
  Dir := ExtractFileDir(FileName);

  if Dir = '' then
  begin
    Result := True;
    Exit
  end;

  p := Pos('\', Dir);
  while p > 0 do
  begin
    SubDir := Copy(Dir, 1, p - 1);
    if not DirectoryExists(SubDir) then
    begin
      Result := CreateDir(SubDir);
      if not Result then
        Exit
    end;
    p := PosEx('\', Dir, p + 1)
  end;

  if not DirectoryExists(Dir) then
    Result := CreateDir(Dir)
  else
    Result := True
end;

function Extract(var PAKFile: file; NodeIdx: Longint;
  FileName: string): Boolean;
var
  Dir    : string;
  OutFile: file;
  Buffer : array of Byte;
begin
  Result := False;

  // convert to local directory name
  FileName := AnsiReplaceStr(FileName, '/', '\');
  Dir := ExtractFilePath(FileName);

  if not MakeDir(Dir) then
  begin
    WriteLn('  error creating/accessing directory');
    Exit
  end;

  FileMode := fmOpenWrite;
  AssignFile(OutFile, FileName);
  {$I-}
  Rewrite(OutFile, 1);
  {$I+}
  if IOResult <> 0 then
  begin
    WriteLn('  error creating output file');
    Exit
  end;

  SetLength(Buffer, HashNodes[NodeIdx].Size);

  {$I-}
  Seek(PAKFile, HashNodes[NodeIdx].Offset);
  BlockRead(PAKFile, Buffer[0], HashNodes[NodeIdx].Size);
  BlockWrite(OutFile, Buffer[0], HashNodes[NodeIdx].Size);
  {$I+}
  if IOResult <> 0 then
    WriteLn('  extraction error');

  Finalize(Buffer);

  CloseFile(OutFile);

  Result := (IOResult = 0)
end;

function ConvFileName(const FileName: string; out ConvName: string): Boolean;
var
  i: Longint;
begin
  Result := True;

  SetLength(ConvName, Length(FileName));

  for i := 1 to Length(FileName) do
    ConvName[i] := Chr(NameTableInv[Ord(FileName[i])]);

  for i := 1 to Length(ConvName) do
    Result := Result and (ConvName[i] <> #000)
end;

function FindNode(ConvName: string; BaseNodeIdx, PathLen: Longint;
  out FileNodeIdx: Longint): Boolean;
var
  i,
  NodeIdx: Longint;
  CurNode: THashNode;
begin
  Result := False;

  if (BaseNodeIdx < 0) or (BaseNodeIdx > High(HashNodes)) then
    Exit;

  if ConvName = '' then
    Exit;

  // hash by next character
  NodeIdx := Ord(ConvName[1]);
  if NodeIdx = 0 then
    Exit;

  Inc(NodeIdx, BaseNodeIdx);
  if NodeIdx > High(HashNodes) then
    Exit;

  CurNode := HashNodes[NodeIdx];

  i := 0;
  while Names[CurNode.NameIdx + i] <> 0 do
  begin
    if i + 2 > Length(ConvName) then
      Exit;
    if Names[CurNode.NameIdx + i] <> Ord(ConvName[i + 2]) then
      Exit;

    Inc(i)
  end;

  Inc(PathLen, i + 1);

  if (CurNode.Size <> 0) and (PathLen = CurNode.PathLen) then
  begin
    FileNodeIdx := NodeIdx;
    Result := True
  end
  else if (CurNode.Size = 0) and (PathLen = CurNode.PathLen + 1) then
    Result := FindNode(RightStr(ConvName, Length(ConvName) - i - 1),
      CurNode.BaseNodeIdx, PathLen, FileNodeIdx)
end;

procedure ExtractAll(var PAKFile: file; var DirFile: TextFile);
var
  FileName,
  ConvName    : string;
  i,
  FileNodeIdx,
  NumFiles,
  NumExtracted: Longint;
begin
  NumFiles := 0;
  NumExtracted := 0;

  for i := 0 to High(HashNodes) do
  if HashNodes[i].Size > 0 then
    Inc(NumFiles);

  while not Eof(DirFile) do
  begin
    ReadLn(DirFile, FileName);

    WriteLn(FileName);

    if not ConvFileName(FileName, ConvName) then
    begin
      WriteLn('  invalid file name');
      Continue
    end;

    if not FindNode(ConvName, HashNodes[0].BaseNodeIdx, 0, FileNodeIdx) then
    begin
      WriteLn('  not found in PAK');
      Continue
    end;

    Extract(PAKFile, FileNodeIdx, FileName);
    Inc(NumExtracted)
  end;

  WriteLn(Format('%d files in PAK, %d extracted.', [NumFiles, NumExtracted]))
end;

procedure ProcessPAK(const PAKName, DirName: string);
var
  PAKFile  : file;
  DirFile  : TextFile;
  PAKHeader: TPAKHeader;
begin
  if not OpenPAKFile(PAKFile, PAKName) then
  begin
    WriteLn(Format('Could not open %s', [PAKName]));
    Exit
  end;

  if not OpenDirFile(DirFile, DirName) then
  begin
    WriteLn(Format('Could not open %s', [DirName]));
    CloseFile(PAKFile);
    Exit
  end;

  if not ReadHeader(PAKFile, PAKHeader) then
  begin
    WriteLn(Format('%s is not a valid PAK file', [PAKName]));
    CloseFile(DirFile);
    CloseFile(PAKFile);
    Exit
  end;

  // read structures
  with PAKHeader do
  begin
    SetLength(HashNodes, NumHashNodes);
    SetLength(Names, NumNames);
    SetLength(NameIdxs, NumNames);
    {$I-}
    BlockRead(PAKFile, HashNodes[0], NumHashNodes * SizeOf(THashNode));
    BlockRead(PAKFile, Names[0], NumNames * SizeOf(Byte));
    BlockRead(PAKFile, NameIdxs[0], NumNameIdxs * SizeOf(Longint))
    {$I+}
  end;

  if IOResult = 0 then
    ExtractAll(PAKFile, DirFile)
  else
    WriteLn('Error reading PAK structres');

  CloseFile(DirFile);
  CloseFile(PAKFile)
end;

begin
  if ParamCount = 2 then
    ProcessPAK(ParamStr(1), ParamStr(2))
  else
    Writeln('Incorrect number of parameters.')
end.

