{ File   : SharkReader.dpr
  Author : Deniz Oezmen
  Created: 2006-10-17
  Changed: 2009-06-19

  Borland Delphi translation and adaptation of the C# sources to "sharkex".
  (Original copyright unknown, possibly a fellow going by the nick of
  "pinkunozou".)
  Adapted to Shark3D ASCII file format, added BMP extractor especially for
  Dreamfall savegames.
}
program SharkReader2;

{$APPTYPE CONSOLE}

uses
  SysUtils,
  StrUtils,
  SaveBMP3;

type
  TColor = packed record
             R, G, B, A: Byte;
           end;

var
  StrDict  : array of record
                        Idx: Longint;
                        Str: string;
                      end;
  StrCtr,
  SX, SY   : Longint;
  HasBitmap: Boolean;

function ReadStringBasic(var InFile: file): string;
var
  c: Char;
begin
  Result := '';

  repeat
    BlockRead(InFile, c, 1);
    Result := Result + c
  until c = #000;

  Result := LeftStr(Result, Length(Result) - 1)
end;

function ContainsKey(Idx: Longint): Boolean;
var
  i: Longint;
begin
  Result := False;

  for i := 0 to High(StrDict) do
    Result := Result or (StrDict[i].Idx = Idx)
end;

function ReadNum(var InFile: file): Longint;
var
  n    : Byte;
  Shift: Longint;
begin
  Shift := 0;
  Result := 0;

  repeat
    BlockRead(InFile, n, 1);
    Result := Result or ((n and $7f) shl Shift);
    Inc(Shift, 7)
  until n < $80
end;

function GetEntry(Idx: Longint): string;
var
  i: Longint;
begin
  Result := '';

  for i := 0 to High(StrDict) do
    if StrDict[i].Idx = Idx then
    begin
      Result := StrDict[i].Str;
      Break
    end
end;

function ReadString(var InFile: file): string;
var
  Num,
  Idx: Longint;
begin
  Num := ReadNum(InFile);
  Idx := StrCtr - Num;

  if Num = 0 then
    Inc(StrCtr);

  if ContainsKey(Idx) then
  begin
    Result := GetEntry(Idx);
    Exit
  end;

  Result := ReadStringBasic(InFile);

  SetLength(StrDict, Length(StrDict) + 1);
  StrDict[High(StrDict)].Idx := Idx;
  StrDict[High(StrDict)].Str := Result
end;

function ReadFloat(var InFile: file): Single;
var
  i: Longint;
  Buf: array[0..3] of Byte;
begin
  for i := 3 downto 0 do
    BlockRead(InFile, Buf[i], 1);
  Move(Buf[0], Result, 4)
end;

procedure ReadSub(var InFile: file; var OutFile: TextFile;
  const Prefix: string);
var
  i, j,
  Num,
  Val,
  n         : Longint;
  Name,
  Line      : string;
  AttachCode: Byte;
begin
  Num := ReadNum(InFile);

  for i := 1 to Num do
  begin
    Name := ReadString(InFile);
    Line := Prefix + Name;

    BlockRead(InFile, AttachCode, 1);

    case AttachCode of
      $00: WriteLn(OutFile, Line);
      $01: begin
             Val := ReadNum(InFile);
             WriteLn(OutFile, Format('%s %u', [Line, Val]));
             if (Name = 'hasbitmap') and (Val = 1) then
               HasBitmap := True
             else if Name = 'sx' then
               SX := Val
             else if Name = 'sy' then
               SY := Val
           end;
      $02: begin
             n := ReadNum(InFile);
             WriteLn(OutFile, Line);
             WriteLn(OutFile, Prefix + '{');
             Line := Prefix + '    ';
             for j := 1 to n do
             begin
               Line := Line + Format('%u', [ReadNum(InFile)]);
               if j < n then
                 Line := Line + ', ';
               if Length(Line) > 70 then
               begin
                 WriteLn(OutFile, Line);
                 Line := Prefix + '    '
               end
             end;
             if Line <> Prefix + '    ' then
               WriteLn(OutFile, Line);
             WriteLn(OutFile, Prefix + '}')
           end;
      $04: WriteLn(OutFile, Format('%s %.7g', [Line, ReadFloat(InFile)]));
      $08: begin
             n := ReadNum(InFile);
             WriteLn(OutFile, Line);
             WriteLn(OutFile, Prefix + '{');
             Line := Prefix + '    ';
             for j := 1 to n do
             begin
               Line := Line + Format('%.7g', [ReadFloat(InFile)]);
               if j < n then
                 Line := Line + ', ';
               if Length(Line) > 70 then
               begin
                 WriteLn(OutFile, Line);
                 Line := Prefix + '    '
               end
             end;
             if Line <> Prefix + '    ' then
               WriteLn(OutFile, Line);
             WriteLn(OutFile, Prefix + '}')
           end;
      $10: WriteLn(OutFile, Format('%s "%s"', [Line, ReadString(InFile)]));
      $20: begin
             n := ReadNum(InFile);
             WriteLn(OutFile, Line);
             WriteLn(OutFile, Prefix + '{');
             for j := 1 to n do
             begin
               Write(OutFile, Format('%s  "%s"',
                 [Prefix, ReadString(InFile)]));
               if j < n then
                 WriteLn(OutFile, ',')
               else
                 WriteLn(OutFile)
             end;
             WriteLn(OutFile, Prefix + '}')
           end;
      $40: begin
             WriteLn(OutFile, Line);
             WriteLn(OutFile, Prefix + '{');
             ReadSub(InFile, OutFile, Prefix + '    ');
             WriteLn(OutFile, Prefix + '}')
           end;
      $80: begin
             WriteLn(OutFile, Line);
             n := ReadNum(InFile);
             for j := 1 to n do
             begin
               WriteLn(OutFile, Prefix + '{');
               ReadSub(InFile, OutFile, Prefix + '    ');
               if j <= n - 1 then
                 WriteLn(OutFile, Prefix + '}')
               else
                 WriteLn(OutFile, Prefix + '}')
             end
           end
    else
      WriteLn(OutFile, Format('Unknown attach code %.2x (aborting)',
        [AttachCode]));
      Exit
    end
  end
end;

procedure ReadShark(var InFile: file; var OutFile: TextFile;
  const Name: string);
var
  Version: string;
  BMP    : TBMPFile;
  x, y   : Longint;
  BMPBuf : array of TColor;
begin
  if ReadStringBasic(InFile) <> 'shark3d_snake_binary' then
    Exit;

  StrCtr := 0;
  SetLength(StrDict, StrCtr);

  Version := ReadStringBasic(InFile);

  DecimalSeparator := '.';
  HasBitmap := False;
  SX := -1;
  SY := -1;

  ReadSub(InFile, OutFile, '');

  // does the file contain a valid bitmap?
  if HasBitmap and (SX > 0) and (SY > 0) and
    (FilePos(InFile) + SX * SY * SizeOf(TColor) <= FileSize(InFile)) then
    begin
      WriteLn('  bitmap found');
      OpenBMP(BMP, Name + '.bmp', SX, SY);
      SetLength(BMPBuf, SX * SY);
      BlockRead(Infile, BMPBuf[0], Length(BMPBuf) * SizeOf(TColor));
      for y := 0 to SY - 1 do
        for x := 0 to SX - 1 do
          with BMPBuf[y * SY + x] do
            WritePixelToBMP(BMP, x, y, R, G, B);
      Finalize(BMPBuf);
      CloseBMP(BMP)
    end;

  Finalize(StrDict)
end;

procedure ScanFiles(FileParam: AnsiString);
var
  FileName,
  Dir,
  Name     : AnsiString;
  InFile   : file;
  OutFile  : TextFile;
  SearchRec: TSearchRec;
begin
  // get the directory information from FileParam
  Dir := ExtractFilePath(FileParam);

  // search for all files matching FileParam
  if FindFirst(FileParam, faAnyFile - faDirectory, SearchRec) = 0 then
  begin
    repeat
      FileName := SearchRec.Name;
      Name := Dir + FileName;

      WriteLn(Name);

      // if this file is accessible, analyze it
      {$I-}
      AssignFile(InFile, Name);
      Reset(InFile, 1);
      {$I+}
      if IOResult = 0 then
      begin
        {$I-}
        AssignFile(OutFile, Name + '.txt');
        Rewrite(OutFile);
        {$I+}
        if IOResult = 0 then
        begin
          ReadShark(InFile, OutFile, Name);
          CloseFile(OutFile)
        end
        else
          WriteLn('  write error');

        CloseFile(InFile)
      end
      else
        WriteLn('  not accessible');

    until FindNext(SearchRec) <> 0
  end
  else
    WriteLn(Format('No matches for "%s"', [FileParam]));

  FindClose(SearchRec)
end;

begin
  if ParamCount = 1 then
    ScanFiles(ParamStr(1))
  else
    WriteLn('Incorrect number of parameters.')
end.

